#!/usr/bin/python
# This script must be run with bin/py.

from lpscripts import (
    AbortException, FinishException, LaunchpadAPIScript, flash_message,
    prompt, reject_projects)
from datetime import date


MAXIMUM_DAYS = 90

class ExpiredCommercialSubscription(LaunchpadAPIScript):
    """Find projects with expired commercial subscriptions and notify them.

    When a commercial subscription expires the project admins are shown a
    message on the project index page.  No other notification is done.  This
    script will hunt them down and send out an email.
    """

    SUBJECT = "Your project in Launchpad (%(project_name)s)"
    FEEDBACK = None

    USAGE="""\
%prog [options] [projects]

'projects' is a list of project names or URLs to projects in
Launchpad which will have their subscription examined and optionally
notified.  If no projects are given then all projects are examined.
"""

    MSG_TEMPLATE = u"""\
Hello %(owner_display_name)s,

Your commercial project '%(project_name)s' in Launchpad
has a commercial use subscription that expired on %(expiration_date)s.

In order to continue using Launchpad for your project (unless you've
relicensed it as open source) you will need to obtain a new subscription by
going to the Canonical shop and purchasing a 'Launchpad Commercial Use
Voucher' and then applying it to your project.  The steps are:

1) Go to https://shop.canonical.com/product_info.php?products_id=230 and
purchase a voucher.

2) Wait a few minutes and go to https://launchpad.net/people/+me/+vouchers and
apply your voucher to your project.  The voucher should show up on that page
within fifteen minutes.

If you've chosen to let your subscription lapse and do not wish to renew,
please let me know and I'll disable your project.  Otherwise it will be
disabled in one week.

Thank you for your past use of Launchpad.  We hope you'll continue to host
your project with us.

Thanks,

%(me)s
Canonical, Ltd.
"""

    WHITEBOARD_MESSAGE = """\
%(today)s %(me)s: Owner %(owner_display_name)s \
(%(owner_name)s) notified of expired license.  Disable after one week."""

    def message_data(self, proj, owner):
        """Construct a dictionary with data for interpolating messages."""
        data = super(ExpiredCommercialSubscription,
                     self).message_data(proj, owner)
        data['expiration_date'] = self.expiration_date
        return data

    def addParserOptions(self):
        super(ExpiredCommercialSubscription, self).addParserOptions()
        self.parser.add_option(
            '-d', '--date',
            default=str(date.today()), dest='expiry_date',
            help="""\
Expiration date to use.  All projects after the date are returned.""")

        self.parser.add_option(
            '--notify',
            action='store_true', default=False, dest='notify',
            help="""Notify the project owner/s via email.""")

    def validateArgs(self):
        """Validate the args parameter.

        Set self.parser.error if a problem is found.
        """
        # Args are optional, so don't do a check.
        pass

    def get_projects(self):
        """Find all expired projects or return the list the user provided."""
        if self.args:
            return self.args

        # Find all projects expired after the given date.
        expired = self.lp.projects.licensing_search(
            subscription_expires_before=self.opts.expiry_date,
            active=True)
        return [proj for proj in expired if not proj.is_permitted]

    def process_project(self, proj):
        subscription = proj.commercial_subscription
        if subscription is None:
            print "%s does not have a commercial subscription." % (
                proj.display_name)
            return
        owner = proj.owner
        # We want to send an email to an individual owner not a team.  Since
        # teams can be the owners of other teams, let's recurse until we find
        # an individual owner.
        while owner.is_team:
            owner = owner.team_owner
        self.expiration_date = subscription.date_expires.date()
        print proj.display_name
        print "Created:", proj.date_created.date()
        print "Expires: %s" %  (
            self.expiration_date,)
        print "Registrant: %s (%s)" % (
            proj.registrant.display_name,
            proj.registrant.preferred_email_address.email)
        print "Maintainer: %s" % (
            proj.owner.display_name)
        print "Individual contact: %s (%s)" % (
            owner.display_name,
            owner.preferred_email_address.email)
        print "Licenses: %s" % ', '.join(proj.licenses)
        wb = proj.reviewer_whiteboard
        if wb:
            print "-----"
            print wb
        print "-----"
        while True:
            char = prompt(
                '(e)mail (d)isable (s)kip (f)inish (q)uit without saving ')
            if char == 'e':
                self.send_email(proj, owner)
                self.update_whiteboard(proj, owner)
                break
            elif char == 'd':
                opts = self.opts
                opts.msg_num = 3
                opts.force = True
                reject_projects(self.lp, [proj], opts=opts)
                break
            elif char == 's':
                break
            elif char == 'f':
                # Save all outstanding actions and quit.
                raise FinishException()
            elif char == 'q':
                # Quit without saving.
                raise AbortException()
            else:
                flash_message("Please select a valid option.")(1)


if __name__ == "__main__":

    script = ExpiredCommercialSubscription()
    script.run()
