#! /usr/bin/python

"""
This script can be used to compute statistics on a Launchpad access log.

The default is to report the number of requests for each <METHOD> <URL>.

The key used for aggregation can be configured through the --key option. For
example --key pageid, will compute the statistics by page id.

The statistics computed is configured through the --stats option. Available
statistics are:

    count       Return the number of requests.
    sum:field   Return the sum of values for <field>.
    avg:field   Return average value for <field> (useful for
                request_duration).
    min:field   Return minimal value for <field>.
    max:field   Return maximal value for <field>.

The --start and --end options can be used to compute the statistics only on a
subset of the log file. Timestamp is specified using yyyy-mm-dd.HH:MM:SS
format. Only the yyyy-mm-dd part is required.

Available keys for aggregation:
    x_forwarded_for     the original client that made the request
    userid              the name of the logged in user
    pageid              the page id
    referrer            the referrer
    referrer_host       the referring host
    referrer_path       the URL path portion of the referrer
    referrer_query      the query string component of the referring URL
    user_agent          the user agent string
    status              the HTTP status result.
    method              the HTTP method
    url                 the URL path requested
    query               the query string of the request
    version             the HTTP version of the request

Available fields to compute the statistics on:

    bytes_written       bytes returned to the client
    sql_statements      number of SQL statements issued
    request_duration    time taken to render the results
    traversal_ticks     Python bytecode executed to traverse to the object.
    publication_ticks   Python bytecode executed to serve the request.
"""

import re
import sys
import pytz
import warnings

from datetime import datetime, timedelta
from operator import itemgetter
from optparse import OptionParser, OptionValueError
from urlparse import urlsplit


TIMESTAMP_RE = re.compile(
    '(\d{1,2})/(\w+)/(\d{4}):(\d{2}):(\d{2}):(\d{2}) ([-+]\d{4})')
Abbrev2Month = [
    'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul',
    'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

FORMAT_SPEC = {
    float: '%10.2f %s',
    int: '%10d %s'}


class InvalidTimestamp(ValueError):
    """An invalid CLF timestamp was encountered."""


class Count:
    """An aggregator counting the number of records."""

    def __init__(self):
        self.count = 0

    def update(self, record):
        self.count +=1

    def finalize(self):
        return self.count


class Sum:
    """An aggregator that sums a particular record field."""

    def __init__(self, field):
        self.field_getter = itemgetter(field)
        self.sum = 0

    def update(self, record):
        self.sum += self.field_getter(record)

    def finalize(self):
        return self.sum

class Average:
    """An aggregator that averages a particular record field."""

    def __init__(self, field):
        self.field_getter = itemgetter(field)
        self.sum = 0
        self.count = 0

    def update(self, record):
        self.count += 1
        self.sum += self.field_getter(record)

    def finalize(self):
        if self.count == 0:
            return 'n/a'
        else:
            return self.sum / float(self.count)


class Max:
    """An aggregator that returns the maximum value for the field."""

    def __init__(self, field):
        self.field_getter = itemgetter(field)
        self.max = None

    def update(self, record):
        value = self.field_getter(record)
        if self.max is None or value > self.max:
            self.max = value

    def finalize(self):
        if self.max is None:
            return 'n/a'
        else:
            return self.max


class Min:
    """An aggregator that returns the minimum value for the field."""

    def __init__(self, field):
        self.field_getter = itemgetter(field)
        self.min = None

    def update(self, record):
        value = self.field_getter(record)
        if self.min is None or value < self.min:
            self.min = value

    def finalize(self):
        if self.min is None:
            return 'n/a'
        else:
            return self.min



def parse_timestamp(timestamp):
    match = TIMESTAMP_RE.search(timestamp)
    if not match:
        raise InvalidTimestamp('invalid CLF timestamp: %s' % timestamp)
    year = int(match.group(3))
    month = Abbrev2Month.index(match.group(2))+1
    day = int(match.group(1))
    hours = int(match.group(4))
    minutes= int(match.group(5))
    seconds = int(match.group(6))
    tz = match.group(7)

    # Normalize to UTC.
    offset = timedelta(hours=int(tz[1:3]), minutes=int(tz[3:]))
    if tz[0] == '+':
        offset = -offset

    return datetime(
        year, month, day, hours, minutes, seconds, tzinfo=pytz.utc) + offset


def parse_access_record(line):
    # log_template = (' - "%s" "%s" [%s] "%s" %s %d %d %s %s '
    #                 '%s "%s" "%s" "%s" "%s"\n')
    #      task.channel.addr[0],
    #      log_template % (
    #          x_forwarded_for,
    #          host,
    #          start_time,
    #          first_line,
    #          status,
    #          bytes_written,
    #          sql_statements,
    #          request_duration,
    #          traversal_ticks,
    #          publication_ticks,
    #          userid,
    #          pageid,
    #          referrer,
    #          user_agent
    #
    record = {
        # Initialize computed fields that might be missing
        'query': '',
        'referrer_host': '',
        'referrer_path': '',
        'referrer_query': '',
        }
    record['client_addr'], ignored, rest = line.split(' ', 2)
    x_forwarded_for, host, rest = line.split('" ', 2)
    record['x_forwarded_for'] = x_forwarded_for[1:]
    record['host'] = host[1:]
    timestamp , rest = rest.split('] ', 1)
    record['timestamp'] = parse_timestamp(timestamp[1:])
    first_line, various_fields, pageid, referrer, user_agent = (
        rest.split('" ', 4))
    record['method'], record['url'], record['version'] = (
        first_line[1:].split(' ', 2))
    if '?' in record['url']:
        record['url'], record['query'] = record['url'].split('?', 1)
    various_fields = various_fields.split(' ', 6)
    record['status'] = int(various_fields[0])
    record['bytes_written'] = int(various_fields[1])
    record['sql_statements'] = int(various_fields[2])
    record['request_duration'] = float(various_fields[3])
    record['traversal_ticks'] = int(various_fields[4])
    record['publication_ticks'] = int(various_fields[5])
    record['userid'] = various_fields[6][1:]
    record['pageid'] = pageid[1:]
    record['referrer'] = referrer = referrer[1:]
    if referrer:
        (address, record['referrer_host'], record['referrer_path'],
            record['referrer_query'], fragment) = urlsplit(record['referrer'])
    record['referrer'] = referrer[1:]
    record['user_agent'] = user_agent[1:]
    return record


def method_path_key(record):
    return "%s %s" % (record['method'], record['url'])


def summarize_by_key(log, keygetter, aggregator_factory=Count):
    """Go over the log, aggregating results for a particular key."""
    summary = {}
    for record in log:
        key = keygetter(record)
        if key not in summary:
            summary[key] = aggregator_factory()
        summary.get(key).update(record)
    return reversed(sorted(
        (aggregator.finalize(), key) for key, aggregator in summary.items()))


class LaunchpadAccessLog:
    """Iterator over the page logs in a file."""

    def __init__(self, fh):
        self.fh = fh


    def __iter__(self):
        """Iterates of the CLF in a access log."""
        for line_no, line in enumerate(self.fh.readlines()):
            try:
                yield parse_access_record(line)
            except ValueError, e:
                warnings.warn("error parsing line %d: %s" % (line_no, e))


class FilteredLog:
    """Filter around records."""

    def __init__(self, log, start_time, end_time):
        self.log = log
        self.start_time = start_time
        self.end_time = end_time

    def __iter__(self):
        """Return records matching the filter."""
        start = self.start_time
        end = self.end_time
        for record in self.log:
            ts = record['timestamp']
            if start and ts < start:
                continue
            if end and ts > end:
                continue
            yield record


def datetime_option(option, opt_str, value, parser):
    """Store an option as a datetime."""
    match = re.match(
        '(\d{4})-(\d{2})-(\d{2}).(\d{2}):(\d{2})(?::(\d{2}))?', value)
    if not match:
        raise OptionValueError(
            'invalid timestamp for %s: %s' % (opt_str, value))
    year = int(match.group(1))
    month = int(match.group(2))
    day = int(match.group(3))
    hours = int(match.group(4))
    minutes = int(match.group(5))
    if match.group(6):
        seconds = int(match.group(6))
    else:
        seconds = 0
    timestamp = datetime(
        year, month, day, hours, minutes, seconds, tzinfo=pytz.utc)
    setattr(parser.values, option.dest, timestamp)


def aggregator_option(option, opt_str, value, parser):
    """Store the aggregator to use based on the option."""
    spec = value.split(':')
    factories = {
        'count': Count,
        'avg': Average,
        'min': Min,
        'max': Max,
        'sum': Sum,
        }
    name, args = spec[0], spec[1:]
    if name not in factories:
        raise OptionValueError('invalid statistics specification: %s' % value)
    setattr(parser.values, option.dest, lambda: factories[name](*args))


def main():
    parser = OptionParser(usage='usage: %prog [options] [filename]')
    parser.add_option(
        '-s', '--start', dest='start', action="callback",
        callback=datetime_option, nargs=1, type="string",
        help="only summarize records newer than START", metavar='START')
    parser.add_option(
        '-e', '--end', dest='end', action="callback",
        callback=datetime_option, nargs=1, type="string",
        help="only summarize records older than END", metavar='END')
    parser.add_option(
        '-k', '--key', dest='key', default="method,url",
        help=("summarize requests using 'key'. If key is 'total', "
              "summarize over the entire log file. Muliple key can be used "
              "by separating them with commas. Defaults to 'method,url'."))
    parser.add_option(
        '--stats', dest='aggregator', default=Count, action="callback",
        callback=aggregator_option, nargs=1, type="string",
        help=("specify the statistics to compute. "
              "Default is to count the number of requests."))
    (options, args) = parser.parse_args()
    if len(args) == 0:
        fh = sys.stdin
    elif len(args) == 1:
        fh = open(args[0])
    else:
        parser.error('only one filename should be given')

    log = LaunchpadAccessLog(fh)
    if options.start or options.end:
        log = FilteredLog(log, options.start, options.end)
    if options.key == 'total':
        key = lambda x: 'Total'
    elif ',' in options.key:
        keys = [itemgetter(key) for key in options.key.split(',')]
        key = lambda x: " ".join(getter(x) for getter in keys)
    else:
        key = itemgetter(options.key)

    for value, key in summarize_by_key(log, key, options.aggregator):
        print FORMAT_SPEC[type(value)] % (value, key)


if __name__ == '__main__':
    main()
