#!/usr/bin/python

__metaclass__ = type
__all__ = [
    'LaunchpadAPIScript',
    'prompt',
    'flash_message',
    'reject_projects',
    'AbortException',
    'FinishException',
    ]


import os
import sys
import time
from email.Header import Header
from email.MIMEText import MIMEText
from email.Utils import formataddr, formatdate
import optparse
import datetime

from bzrlib.msgeditor import edit_commit_message
from bzrlib.osutils import getchar
from bzrlib.smtp_connection import SMTPConnection
import bzrlib.config

import _pythonpath
from lazr.restfulclient.errors import HTTPError
from launchpadlib.launchpad import Launchpad
import launchpadlib.uris


def valid_service_roots():
    return sorted(launchpadlib.uris.service_roots.keys())


def lp_factory(system_name, app_name):
    """Get a launchpad API instance for ``system_name``."""
    try:
        return Launchpad.login_with(app_name, system_name, version='devel')
    except KeyError:
        print >> sys.stderr, "System '%s' not supported." % system_name
        print >> sys.stderr, "Use one of: ", valid_service_roots()
        return None


def format_address(name, address):
    r"""Formats a name and address to be used as an email header.

        >>> format_address('Name', 'foo@bar.com')
        'Name <foo@bar.com>'
        >>> format_address('', 'foo@bar.com')
        'foo@bar.com'
        >>> format_address(None, u'foo@bar.com')
        'foo@bar.com'

    It handles unicode and characters that need quoting as well.

        >>> format_address(u'F\xf4\xf4 Bar', 'foo.bar@canonical.com')
        '=?utf-8?b?RsO0w7QgQmFy?= <foo.bar@canonical.com>'

        >>> format_address('Foo [Baz] Bar', 'foo.bar@canonical.com')
        '"Foo \\[Baz\\] Bar" <foo.bar@canonical.com>'

    Really long names doesn't get folded, since we're not constructing
    an e-mail header here.

        >>> formatted_address = format_address(
        ...     'a '*100, 'long.name@example.com')
        >>> '\n' in formatted_address
        False
    """
    if not name:
        return str(address)
    name = str(Header(name))
    # Using Header to encode the name has the side-effect that long
    # names are folded, so let's unfold it again.
    name = ''.join(name.splitlines())
    return str(formataddr((name, address)))


class AbortException(Exception):
    """Raised in process_project when the script is to be aborted.

    No outstanding data is to be saved.
    """
    pass


class FinishException(Exception):
    """Raised in process_project when the script is to be aborted.

    Previously accomplished work should be saved.
    """
    pass


class LaunchpadAPIScript:
    """Class for scripts that perform LP admin tasks via launchpadlib."""
    REGISTRY_TEAM = u'registry'

    RTM_TEMPLATE = u"""\
Due: %(due)s
Tags: %(tags)s
URL: %(project_url)s
List: %(list)s
---
%(now)s
%(message)s
"""

    RTM_ADDRESS = 'launchpad+a7141e@rmilk.com'

    RTM_MESSAGE = u""

    RTM_SUBJECT = ""

    USAGE = """\
%prog [options] [projects]

'projects' is a list of project names or URLs to projects in
Launchpad.  This script does not do anything.  Please extend the class
LaunchpadAPIScript to do something useful.
"""

    MSG_TEMPLATE = None
    SUBJECT = ""
    FEEDBACK = 'feedback@launchpad.net'
    COMMASPACE = ', '
    NEWLINE = '\n'

    def message_data(self, proj, owner):
        """Construct a dictionary with data for interpolating messages."""
        project_url = "https://launchpad.net/%s" % proj.name
        return dict(
            owner_name=owner.name,
            owner_display_name=owner.display_name,
            me=self.lp.me.display_name,
            project_url=project_url,
            project_name=proj.name,
            now=datetime.datetime.now(),
            today=datetime.date.today(),
            )

    def send_email_message(self, msg_text, subject, to, sender, cc):
        """Actually send an email message."""
        msg = MIMEText(msg_text.encode('utf-8'), 'plain', 'utf-8')
        msg['From'] = sender
        if self.FEEDBACK:
            msg['Reply-To'] = self.FEEDBACK
        to_list = set()
        if self.opts.to_me:
            msg['To'] = sender
            to_list.add(sender)
        else:
            msg['To'] = to
            to_list.add(to)
            if cc is not None:
                msg['Cc'] = self.COMMASPACE.join(cc)
                to_list.update(cc)
        msg['Subject'] = subject
        msg['Date'] = formatdate()
        msg.epilogue = ''

        message = msg.as_string()

        if self.opts.dry_run:
            print message
            print "Message not sent"
        else:
            print "Sending..."
            # XXX sinzui 2012-01-17: The config API has changed in latest
            # bzr. smtp_connection.py calls config.get('smtp_server')
            # AttributeError: 'GlobalConfig' object has no attribute 'get'
            global_config = bzrlib.config.GlobalConfig()
            if getattr(global_config, 'get', None) is None:
                config = dict(
                    smtp_username=global_config.get_user_option(
                        'smtp_username'),
                    smtp_password=global_config.get_user_option(
                        'smtp_password'),
                    smtp_server=global_config.get_user_option(
                        'smtp_server'),
                    )
            else:
                config = global_config
            SMTPConnection(config).send_email(msg)
            print "email to: ", self.COMMASPACE.join(to_list)

    def send_email(self, project, owner):
        """Send the message to project owner, me, and list."""
        sender = format_address(
            self.lp.me.display_name,
            self.lp.me.preferred_email_address.email)
        to = format_address(
            owner.display_name,
            owner.preferred_email_address.email)
        cc = [sender]
        if self.FEEDBACK:
            cc.append(self.FEEDBACK)
        message_data = self.message_data(project, owner)
        subject = self.SUBJECT % message_data
        text = self.MSG_TEMPLATE % message_data

        if not self.opts.rtm_only:
            if self.opts.edit_msg:
                infotext = (
                    "Edit your outgoing message.  "
                    "Everything above the line will be included.")
                separator = '-' * 27 + ' Do not remove this line ' + 27 * '-'
                try:
                    text = edit_commit_message(
                        infotext=infotext,
                        ignoreline=separator,
                        start_message=text)
                except (KeyboardInterrupt, SystemExit):
                    raise
            self.send_email_message(text, subject, to, sender, None)

        # Log a reminder in RTM via the email interface.
        # opts.rtm_address may be '' or None.
        if self.opts.rtm_address:
            to = self.opts.rtm_address
            subject = self.RTM_SUBJECT % self.message_data(project, owner)
            if self.opts.rtm_only:
                self.opts.rtm_tags += ' +user-not-contacted'
            message_data.update(
                dict(
                    due=self.opts.due,
                    tags=self.opts.rtm_tags,
                    list=self.opts.rtm_list,
                    message=self.rtm_message(project, owner)
                    ))
            text = self.RTM_TEMPLATE % message_data
            self.send_email_message(text, subject, to, sender, None)

    def rtm_message(self, project, owner):
        return self.RTM_MESSAGE % self.message_data(project, owner)

    def whiteboard_message(self, proj, owner):
        return self.WHITEBOARD_MESSAGE % self.message_data(proj, owner)

    def update_whiteboard(self, proj, owner):
        """Make a note on the project whiteboard.

        Record the fact that the owner was notified and the date.
        """
        if self.opts.to_me is True:
            return
        msg = self.whiteboard_message(proj, owner)
        if msg is None:
            return
        old_wb = proj.reviewer_whiteboard
        if old_wb:
            new_wb = self.NEWLINE.join((old_wb, msg))
        else:
            new_wb = msg
        if self.opts.dry_run:
            print "New whiteboard would be:"
            print new_wb
        else:
            proj.reviewer_whiteboard = new_wb
            try:
                proj.lp_save()
            except HTTPError:
                print ("Unable to save new reviewer whiteboard.  "
                       "Please manually add the following:")
                print msg

    def addParserOptions(self):
        """Add arguments to the parser."""
        self.parser.add_option(
            '-s', '--system',
            type='string', default='production', dest='lpsystem',
            help="""\
The Launchpad system to use.  Must be one of %s.""" % valid_service_roots())

        self.parser.add_option(
            '--dryrun',
            action='store_true', default=False, dest='dry_run',
            help="""If this option is used no email will be sent.""")

        self.parser.add_option(
            '--onlyme',
            action='store_true', default=False, dest='to_me',
            help="""\
If this option is used the email is only sent to yourself as a test.""")

        self.parser.add_option(
            '--rtm-only',
            action='store_true', default=False, dest='rtm_only',
            help="""\
Only set a reminder in RTM with the tag +user-not-contacted.
Do not send email.""")

        self.parser.add_option(
            '--rtm',
            type='string', default=None,
            dest='rtm_address',
            help="""\
Provide your RememberTheMilk.com address for emailing tasks.
It can be found under the 'Settings->Info' tab in RTM.
The default is to use the Launchpad group account.""")

        self.parser.add_option(
            '--edit-msg',
            action='store_true', default=False, dest='edit_msg',
            help="""\
Open outgoing message in your EDITOR for customization.""")

    def validateArgs(self):
        """Validate the args parameter.

        Set self.parser.error if a problem is found.
        """
        if len(self.args) < 1:
            self.parser.error('Missing projects')

    def parseArgs(self, args=None):
        """Parse the arguments.

        Sets self.opts and self.args.
        """
        self.parser = optparse.OptionParser(self.USAGE,
                                            conflict_handler="resolve")

        self.addParserOptions()

        if args is None:
            args = sys.argv[1:]
        self.opts, self.args = self.parser.parse_args(args)
        self.validateArgs()

    def normalizeProjectName(self, project_name):
        """Project names may be passed as the full URL.

        Return just the name.
        """
        project_name = project_name.strip()
        if project_name.endswith("/+review"):
            project_name = "/".join(project_name.split("/")[0:-1])
        if (project_name.startswith("https://") or
           project_name.startswith("http://")):
            project_name = project_name.split("/")[-1]
        return project_name

    def process_project(self, project):
        """Process an individual project."""
        pass

    def get_projects(self):
        projects = []
        for project_name in self.args:
            # For convenience, accept URLs as project names.
            if isinstance(project_name, basestring):
                project_name = self.normalizeProjectName(project_name)
                if not project_name:
                    continue
            try:
                proj = self.lp.projects[project_name]
            except KeyError:
                print "%(name)s - no project exists with that name" % dict(
                    name=project_name)
                continue
            projects.append(proj)
        return projects

    def abort(self):
        pass

    def finish(self):
        pass

    def run(self):
        """Process the projects."""
        print "Connecting..."
        sys.stdout.flush()
        try:
            if self.lp is None:
                self.lp = lp_factory(self.opts.lpsystem, self.name)
        except HTTPError, e:
            print "An error has occured contacting Launchpad."
            print e
            sys.exit(-1)
        print "done."

        try:
            for project in self.get_projects():
                print "Examining project %s..." % project.name
                self.process_project(project)
        except FinishException:
            self.finish()
        except AbortException:
            self.abort()

    def __init__(self, name=None, lp=None, args=None, opts=None):
        if name is None:
            name = os.path.split(sys.argv[0])[-1]
        self.name = name
        self.lp = lp
        if opts is None:
            self.parseArgs(args)
        else:
            self.opts = opts


def prompt(message):
    # XXX: Stolen from bzrlib.shelf_ui.
    """Prompt the user for a character.

    :param message: The message to prompt a user with.
    :return: A character.
    """
    sys.stdout.write(message)
    char = getchar()
    sys.stdout.write("\r" + ' ' * len(message) + '\r')
    sys.stdout.flush()
    return char


def flash_message(message):
    """Display a message for a brief time.

    Returns a callable that will clear the message, with an optional
    delay.
    """
    sys.stdout.write(message)
    sys.stdout.flush()

    def clear(delay=0):
        time.sleep(delay)
        sys.stdout.write("\r" + ' ' * len(message) + '\r')
        sys.stdout.flush()

    return clear


def reject_projects(lp, projects, opts=None):
    from disable_projects import DisableProject

    disabler = DisableProject(lp=lp, args=['dummy'], opts=opts)

    for project in projects:
        sys.stdout.write('Rejecting %s...' % project.name)
        disabler.process_project(project)
        try:
            project.lp_save()
        except HTTPError, e:
            reason = e.content.split('\n\n')[0]
            print "  Error saving rejected project: ", reason
        print 'done.'


if __name__ == "__main__":
    script = LaunchpadAPIScript()
    sys.exit()
    script.run()
